=begin

Permission to use, copy, modify, and distribute this software for 
any purpose and without fee is hereby granted, provided that the above
copyright notice appear in all copies.

THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR
IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.

Name:         ColladaWithTextureNames.rb
Author:       Aerilius
Usage:        menu 'File > Export Collada with texture names'

Description:  This plugin allows to export the whole model or the selection 
              to collada fileformat while preserving/restoring the materials original names. 
              This is useful since SketchUp's Collada exporter creates generic names.
              SketchUp also creates new textures images for each face with distorted texture coordinates; 
              however these won't receive the original material's name.
              If you need specific export options (from SketchUp's options dialog), 
              feel free to edit the options_hash below.
Version:      1.2b
Date:         30.09.2011



=end
require 'sketchup.rb'



module ColladaWithTextureNames



def self.run
  @model = Sketchup.active_model
  # ask for filename
  suggested_name = File.basename(@model.path).sub(/^$/,"Unnamed.dae").sub(/\.skp$/,".dae")
  dae_path = UI.savepanel("Export Collada file and preserve texture names", "#{ENV['USERPROFILE']}", suggested_name)
  return nil if dae_path.nil?
  dae_path = dae_path.tr("\\", "/")
  dae_path = dae_path.sub(/(:?.dae)?$/,".dae") # if file extension has been missed
  dae_name = File.basename(dae_path).unpack("U*").map{|c|c.chr}.join.sub(/\.\w{3}$/, "")
  dae_dir = File.join(File.dirname(dae_path), dae_name)
  # export collada
  options_hash = { :triangulated_faces =>  true,
                     :doublesided_faces =>  true,
                     :edges =>  false,
                     :materials_by_layer =>  false,
                     :author_attribution =>  true,
                     :texture_maps =>  true,
                     :selectionset_only =>  false,
                     :preserve_instancing =>  false 
  }
  options_hash[:selectionset_only] = true if !@model.selection.empty?
  @model.export(dae_path, options_hash)
  # export textures into temp directory 
  # (maybe not ideal,
  # but the only way to assign SketchUp's generic texture names to the material's orginal names)
  @materials = {} # hash to remember the materials' screen names
  tw = Sketchup.create_texture_writer
  @model.materials.each{|x|
    next if x.materialType == 0
    @materials[File.basename(x.texture.filename)] = x.name if x.name[/^Material\d+$/].nil?
    tmp_group = @model.entities.add_group
    tmp_group.material = x
    tw.load(tmp_group)
  }
  ref_dir = "#{( ENV['TMPDIR'] || ENV['TMP'] || ENV['TEMP'] ).gsub('\\','/')}/skp_cwtn"
  Dir.mkdir(ref_dir) if !File.exists?(ref_dir)
  # Note: write_all generates unique filenames for different materials that have the same texture filename
  tw.write_all(ref_dir, false)
  # rename textures
  self.rename_textures(dae_path, dae_dir, ref_dir)
  Dir.foreach(ref_dir){|f|
    next if f=='.' or f=='..'
    File.delete(File.join(ref_dir, f))
  } if Dir.entries(ref_dir)[2]
  Dir.delete(ref_dir)
end #def run



##
# Rename textures
# - compares texture files in the collada's supplement directory with texture files in a reference directory
# - renames identical texture files in the collada directory
# - edits the collada file and replaces filenames
#
def self.rename_textures(dae, dir, ref_dir)
  fileNames = {} # hash to remember file name assignment
  materials = {} # hash to remember material screen name assignment, like @materials but with relative paths
  Dir.foreach(dir){|f|
    next if f == "."||f == ".."
    # check reference directory for an identical texture, if so, take its filename
    refFileName = self.file_duplicate?(File.join(dir, f), ref_dir)
    # use either material's screen name, texture's original filename, or don't rename the file
    string = refFileName
    if string && @materials[string]
      string = @materials[string] # material's screen name
    elsif string.nil?
      string = f
    end
    # generate proper, unique filename
    newFileName = self.to_filename(dir, string, File.extname(f))
    File.rename(File.join(dir, f), newFileName)
    # remember filename assignment
    relOld = self.rel_path(File.join(dir, f), dae)
    relNew = self.rel_path(newFileName, dae)
    fileNames[relOld] = relNew
    materials[relOld] = @materials[refFileName] || refFileName
  }
  # read Collada file and replace filenames
  colladaContent = IO.read(dae)
  fileNames.each{|old,new|
    # replace texture.jpg
    colladaContent.gsub!(old,new)
    # search for material screen name and replace it
    imageid = colladaContent.scan(/<image id="([^"]*?)">\s*?<init_from>#{new.gsub('\/','\\\/')}<\/init_from>\s*?<\/image>/m).flatten[0]
    effectid = colladaContent.scan(/<effect id="([^"]*?)">[^\/]*?<init_from>#{imageid}<\/init_from>/m).flatten[0]
    materialid = colladaContent.scan(/<material id="([^"]*?)" name="[^"]*?">\s*?<instance_effect url="##{effectid}"/m).flatten[0]
    materialsymbol = colladaContent.scan(/<instance_material symbol="([^"]*?)" target="##{materialid}"/m).flatten[0]
    colladaContent.gsub!(/(<material id="#{materialid}" name=")[^"]*/m, "\\1"+materials[old]) if materials[old]
    colladaContent.gsub!(/(<instance_material symbol=")[^"]*?(?=" target="##{materialid}">)/m, "\\1"+materials[old]) if materials[old]
    colladaContent.gsub!(/(<triangles count="[^"]" material=")#{materialsymbol}(?=">)/m, "\\1"+materials[old]) if materials[old]
    #colladaContent.gsub!(/material="#{materialsymbol}"/m, "material=\""+materials[old]+"\"") if materials[old]
    #colladaContent.gsub!(/symbol="#{materialsymbol}"/m, "material=\""+materials[old]+"\"") if materials[old]
  }
  colladaFile = File.open(dae, "w")
  colladaFile.write(colladaContent)
  colladaFile.close
  return true
end #def rename_textures



##
# Check if <directory> contains a file that is identical to <file>
#
def self.file_duplicate?(file, directory)
  Dir.foreach(directory){|f|
    return f if File.size(File.join(directory,f)) == File.size(file)
  }
end #def file_duplicate?



##
# Convert a random string into a proper, unique filename
# arguments:
#   dir: directory where the file should go
#   string: any user input
#   ext: force a specific extension if the string doesn't contain one
# returns: a complete file path
# 
def self.to_filename(dir, string, ext=nil)
  # separate extension from basename
  ext = string[/\.\w{1,3}$/].to_s if ext.nil?
  string.sub!(/#{ext}$/,"")
  # clean up basename
  string = string[/.{1,30}/].gsub(/\n|\r/,"").gsub(/[^0-9a-zA-Z\-\_\.]+/,"_").gsub(/^_|_$/,"").to_s
  string = "file" if string.empty?
  base = File.join(dir, string)
  # detect collision of filenames and return alternative filename (numbered)
  base2 = base
  i = 0
  while File.exists?(base2 + ext)
    base2 = base + i.to_s
    i += 1
  end
  return base2 + ext
end



##
# Convert absolute file path into relative path, relative to
# - <base> directory or
# - next common directory with <file> given as second argument
#
def self.rel_path(abs,base)
  base=base.sub(/\/?$/,"/")
  i=0
  i+=1 until abs[i]!=base[i]
  i-=1 if i>0
  rel = abs.scan(/^.{0,#{i}}\/(.*$)/).to_s
  return (i==0)? abs : rel
end #def rel_path



end #module ColladaWithTextureNames



if( not file_loaded?(File.basename(__FILE__)) )
	UI.menu("File").add_item("Export Collada with texture names"){ ColladaWithTextureNames.run() }
end

file_loaded(File.basename(__FILE__))
